@@ -291,13 +291,19 @@ def read_file_bytes(raw_path: str) -> bytes:
         return file.read()
 
 def write_file_bytes(
-        raw_path: str, contents: typing.Union[bytes, None],
+        raw_path: str, contents: typing.Union[bytes, str, None],
         no_clobber: bool = False) -> None:
     """
     Write the contents of a file as bytes.
     If clobbering, any existing dirent will be removed before write.
     """
 
+    if (contents is None):
+        contents = b''
+
+    if (isinstance(contents, str)):
+        contents = contents.encode(DEFAULT_ENCODING)
+
     path = os.path.abspath(raw_path)
 
     if (exists(path)):
@@ -306,9 +312,6 @@ def write_file_bytes(
 
         remove(path)
 
-    if (contents is None):
-        contents = b''
-
     with open(path, 'wb') as file:
         file.write(contents)
 
