@@ -19,6 +19,8 @@ import requests
 import requests_toolbelt.multipart.decoder
 
 import edq.util.dirent
+import edq.util.encoding
+import edq.util.hash
 import edq.util.json
 import edq.util.pyimport
 
@@ -30,6 +32,9 @@ DEFAULT_REQUEST_TIMEOUT_SECS: float = 10.0
 
 DEFAULT_HTTP_EXCHANGE_EXTENSION: str= '.httpex.json'
 
+QUERY_CLIP_LENGTH: int = 100
+""" If the filename of an HTTPExhange being saved is longer than this, then clip it. """
+
 ANCHOR_HEADER_KEY: str = 'edq-anchor'
 """
 By default, requests made via make_request() will send a header with this key that includes the anchor component of the URL.
@@ -37,9 +42,13 @@ Anchors are not traditionally sent in requests, but this will allow exchanges to
 """
 
 ALLOWED_METHODS: typing.List[str] = [
+    'DELETE',
     'GET',
+    'HEAD',
+    'OPTIONS',
+    'PATCH',
     'POST',
-    'PUT'
+    'PUT',
 ]
 """ Allowed HTTP methods for an HTTPExchange. """
 
@@ -58,6 +67,7 @@ DEFAULT_EXCHANGE_IGNORE_HEADERS: typing.List[str] = [
     'etag',
     'host',
     'link',
+    'location',
     'priority',
     'referrer-policy',
     'sec-fetch-dest',
@@ -117,6 +127,7 @@ class FileInfo(edq.util.json.DictConverter):
             path: typing.Union[str, None] = None,
             name: typing.Union[str, None] = None,
             content: typing.Union[str, bytes, None] = None,
+            b64_encoded: bool = False,
             **kwargs: typing.Any) -> None:
         # Normalize the path from POSIX-style to the system's style.
         if (path is not None):
@@ -137,20 +148,45 @@ class FileInfo(edq.util.json.DictConverter):
         self.content: typing.Union[str, bytes, None] = content
         """ The contents of this file. """
 
+        self.b64_encoded: bool = b64_encoded
+        """ Whether the content is a string encoded in Base64. """
+
         if ((self.path is None) and (self.content is None)):
             raise ValueError("File must have either path or content specified.")
 
-    def resolve_path(self, base_dir: str) -> None:
+    def resolve_path(self, base_dir: str, load_file: bool = True) -> None:
         """ Resolve this path relative to the given base dir. """
 
-        if ((self.path is None) or os.path.isabs(self.path)):
-            return
+        if ((self.path is not None) and (not os.path.isabs(self.path))):
+            self.path = os.path.abspath(os.path.join(base_dir, self.path))
 
-        self.path = os.path.abspath(os.path.join(base_dir, self.path))
+        if ((self.path is not None) and (self.content is None) and load_file):
+            self.content = edq.util.dirent.read_file_bytes(self.path)
+
+    def hash_content(self) -> str:
+        """
+        Compute a hash for the content present.
+        If no content is provided, use the path.
+        """
+
+        hash_content = self.content
+
+        if (self.b64_encoded and isinstance(hash_content, str)):
+            hash_content = edq.util.encoding.from_base64(hash_content)
+
+        if (hash_content is None):
+            hash_content = self.path
+
+        return edq.util.hash.sha256_hex(hash_content)
 
     def to_dict(self) -> typing.Dict[str, typing.Any]:
         data = vars(self).copy()
-        del data['content']
+
+        # JSON does not support raw bytes, so we will need to base64 encode any binary content.
+        if (isinstance(self.content, bytes)):
+            data['content'] = edq.util.encoding.to_base64(self.content)
+            data['b64_encoded'] = True
+
         return data
 
     @classmethod
@@ -168,13 +204,13 @@ class HTTPExchange(edq.util.json.DictConverter):
             url_path: typing.Union[str, None] = None,
             url_anchor: typing.Union[str, None] = None,
             parameters: typing.Union[typing.Dict[str, typing.Any], None] = None,
-            files: typing.Union[typing.List[typing.Union[FileInfo, typing.Dict[str, str]]], None] = None,
+            files: typing.Union[typing.List[typing.Union[FileInfo, typing.Dict[str, typing.Any]]], None] = None,
             headers: typing.Union[typing.Dict[str, typing.Any], None] = None,
+            allow_redirects: typing.Union[bool, None] = None,
             response_code: int = http.HTTPStatus.OK,
             response_headers: typing.Union[typing.Dict[str, typing.Any], None] = None,
             json_body: typing.Union[bool, None] = None,
             response_body: typing.Union[str, dict, list, None] = None,
-            read_write: bool = False,
             source_path: typing.Union[str, None] = None,
             response_modifier: typing.Union[str, None] = None,
             extra_options: typing.Union[typing.Dict[str, typing.Any], None] = None,
@@ -235,6 +271,12 @@ class HTTPExchange(edq.util.json.DictConverter):
         self.headers: typing.Dict[str, typing.Any] = headers
         """ Headers in the request. """
 
+        if (allow_redirects is None):
+            allow_redirects = True
+
+        self.allow_redirects: bool = allow_redirects
+        """ Follow redirects. """
+
         self.response_code: int = response_code
         """ The HTTP status code of the response. """
 
@@ -262,13 +304,6 @@ class HTTPExchange(edq.util.json.DictConverter):
         The response that should be sent in this exchange.
         """
 
-        self.read_write: bool = read_write
-        """
-        Indicates that this exchange will change data on the server (regardless of the HTTP method).
-        This field may be ignored by test servers,
-        but may be observed by tools that generate or validate test data.
-        """
-
         self.response_modifier: typing.Union[str, None] = response_modifier
         """
         This function reference will be used to modify responses (in HTTPExchange.make_request() and HTTPExchange.from_response())
@@ -316,7 +351,7 @@ class HTTPExchange(edq.util.json.DictConverter):
         if (url_path is not None):
             url_path = url_path.strip()
             if (url_path == ''):
-                url_path = None
+                url_path = ''
             else:
                 url_path = url_path.lstrip('/')
 
@@ -414,9 +449,9 @@ class HTTPExchange(edq.util.json.DictConverter):
         if (not match):
             return False, hint
 
-        # Check file names, not file content.
-        query_filenames = {file.name for file in query.files}
-        target_filenames = {file.name for file in self.files}
+        # Check file names and hash contents.
+        query_filenames = {(file.name, file.hash_content()) for file in query.files}
+        target_filenames = {(file.name, file.hash_content()) for file in self.files}
         if (query_filenames != target_filenames):
             return False, f"File names do not match (query = {query_filenames}, target = {target_filenames})."
 
@@ -472,6 +507,12 @@ class HTTPExchange(edq.util.json.DictConverter):
         files = []
         for file_info in self.files:
             content = file_info.content
+
+            # Content is base64 encoded.
+            if (file_info.b64_encoded and isinstance(content, str)):
+                content = edq.util.encoding.from_base64(content)
+
+            # Content is missing and must be in a file.
             if (content is None):
                 content = open(file_info.path, 'rb')  # type: ignore[assignment,arg-type]  # pylint: disable=consider-using-with
 
@@ -484,6 +525,7 @@ class HTTPExchange(edq.util.json.DictConverter):
                 data = self.parameters,
                 files = files,
                 raise_for_status = raise_for_status,
+                allow_redirects = self.allow_redirects,
                 **kwargs,
         )
 
@@ -543,6 +585,44 @@ class HTTPExchange(edq.util.json.DictConverter):
 
         return True, None
 
+    def compute_relpath(self, http_exchange_extension: str = DEFAULT_HTTP_EXCHANGE_EXTENSION) -> str:
+        """ Create a consistent, semi-unique, and relative path for this exchange. """
+
+        url = self.get_url().strip()
+        parts = url.split('/')
+
+
+        if (url in ['', '/']):
+            filename = '_index_'
+            dirname = ''
+        else:
+            filename = parts[-1]
+
+            if (len(parts) > 1):
+                dirname = os.path.join(*parts[0:-1])
+            else:
+                dirname = ''
+
+        parameters = {}
+        for key in sorted(self.parameters.keys()):
+            parameters[key] = self.parameters[key]
+
+        # Treat files as params as well.
+        for file_info in self.files:
+            parameters[f"file-{file_info.name}"] = file_info.hash_content()
+
+        query = urllib.parse.urlencode(parameters)
+        if (query != ''):
+            # The query can get very long, so we may have to clip it.
+            query_text = edq.util.hash.clip_text(query, QUERY_CLIP_LENGTH)
+
+            # Note that the '?' is URL encoded.
+            filename += f"%3F{query_text}"
+
+        filename += f"_{self.method}{http_exchange_extension}"
+
+        return os.path.join(dirname, filename)
+
     def to_dict(self) -> typing.Dict[str, typing.Any]:
         return vars(self)
 
@@ -573,6 +653,7 @@ class HTTPExchange(edq.util.json.DictConverter):
             response: requests.Response,
             headers_to_skip: typing.Union[typing.List[str], None] = None,
             params_to_skip: typing.Union[typing.List[str], None] = None,
+            allow_redirects: typing.Union[bool, None] = None,
             ) -> 'HTTPExchange':
         """ Create a full excahnge from a response. """
 
@@ -618,10 +699,27 @@ class HTTPExchange(edq.util.json.DictConverter):
             'response_headers': response_headers,
             'response_body': body,
             'response_modifier': _exchanges_clean_func,
+            'allow_redirects': allow_redirects,
         }
 
         return HTTPExchange(**data)
 
+@typing.runtime_checkable
+class HTTPExchangeComplete(typing.Protocol):
+    """
+    A function that can be called after a request has been made (and exchange constructed).
+    """
+
+    def __call__(self,
+            exchange: HTTPExchange
+            ) -> str:
+        """
+        Called after an HTTP exchange has been completed.
+        """
+
+_make_request_exchange_complete_func: typing.Union[HTTPExchangeComplete, None] = None  # pylint: disable=invalid-name
+""" If not None, call this func after make_request() has created its HTTPExchange. """
+
 def find_open_port(
         start_port: int = DEFAULT_START_PORT, end_port: int = DEFAULT_END_PORT,
         wait_time: float = DEFAULT_PORT_SEARCH_WAIT_SEC) -> int:
@@ -666,6 +764,9 @@ def make_request(method: str, url: str,
         params_to_skip: typing.Union[typing.List[str], None] = None,
         http_exchange_extension: str = DEFAULT_HTTP_EXCHANGE_EXTENSION,
         add_http_prefix: bool = True,
+        additional_requests_options: typing.Union[typing.Dict[str, typing.Any], None] = None,
+        exchange_complete_func: typing.Union[HTTPExchangeComplete, None] = None,
+        allow_redirects: typing.Union[bool, None] = None,
         **kwargs: typing.Any) -> typing.Tuple[requests.Response, str]:
     """
     Make an HTTP request and return the response object and text body.
@@ -686,6 +787,9 @@ def make_request(method: str, url: str,
     if (files is None):
         files = []
 
+    if (additional_requests_options is None):
+        additional_requests_options = {}
+
     # Add in the anchor as a header (since it is not traditionally sent in an HTTP request).
     if (send_anchor_header):
         headers = headers.copy()
@@ -693,18 +797,22 @@ def make_request(method: str, url: str,
         parts = urllib.parse.urlparse(url)
         headers[ANCHOR_HEADER_KEY] = parts.fragment.lstrip('#')
 
-    options = {
+    options = additional_requests_options.copy()
+    options.update({
         'headers': headers,
         'files': files,
         'timeout': timeout_secs,
-    }
+    })
+
+    if (allow_redirects is not None):
+        options['allow_redirects'] = allow_redirects
 
     if (method == 'GET'):
         options['params'] = data
     else:
         options['data'] = data
 
-    logging.debug("Making %s request: '%s'.", method, url)
+    logging.debug("Making %s request: '%s' (options = %s).", method, url, options)
     response = requests.request(method, url, **options)
 
     body = response.text
@@ -717,21 +825,25 @@ def make_request(method: str, url: str,
 
         response.raise_for_status()
 
-    if (output_dir is not None):
-        exchange = HTTPExchange.from_response(response, headers_to_skip = headers_to_skip, params_to_skip = params_to_skip)
+    exchange = None
+    if ((output_dir is not None) or (exchange_complete_func is not None) or (_make_request_exchange_complete_func is not None)):
+        exchange = HTTPExchange.from_response(response,
+                headers_to_skip = headers_to_skip, params_to_skip = params_to_skip,
+                allow_redirects = options.get('allow_redirects', None))
 
-        path = os.path.abspath(os.path.join(output_dir, *exchange.get_url().split('/')))
-
-        query = urllib.parse.urlencode(exchange.parameters)
-        if (query != ''):
-            # Note that the '?' is URL encoded.
-            path += f"%3F{query}"
-
-        path += f"_{method}{http_exchange_extension}"
+    if ((output_dir is not None) and (exchange is not None)):
+        relpath = exchange.compute_relpath(http_exchange_extension = http_exchange_extension)
+        path = os.path.abspath(os.path.join(output_dir, relpath))
 
         edq.util.dirent.mkdir(os.path.dirname(path))
         edq.util.json.dump_path(exchange, path, indent = 4, sort_keys = False)
 
+    if ((exchange_complete_func is not None) and (exchange is not None)):
+        exchange_complete_func(exchange)
+
+    if ((_make_request_exchange_complete_func is not None) and (exchange is not None)):
+        _make_request_exchange_complete_func(exchange)  # pylint: disable=not-callable
+
     return response, body
 
 def make_get(url: str, **kwargs: typing.Any) -> typing.Tuple[requests.Response, str]:
@@ -850,13 +962,14 @@ def parse_content_dispositions(headers: typing.Union[email.message.Message, typi
 
 def parse_query_string(text: str,
         replace_single_lists: bool = True,
-        ) -> typing.Dict[str, typing.Any]:
+        keep_blank_values: bool = True,
+        **kwargs: typing.Any) -> typing.Dict[str, typing.Any]:
     """
     Parse a query string (like urllib.parse.parse_qs()), and normalize the result.
     If specified, lists with single values (as returned from urllib.parse.parse_qs()) will be replaced with the single value.
     """
 
-    results = urllib.parse.parse_qs(text)
+    results = urllib.parse.parse_qs(text, keep_blank_values = True)
     for (key, value) in results.items():
         if (replace_single_lists and (len(value) == 1)):
             results[key] = value[0]  # type: ignore[assignment]
